// Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Ingmar Stapel: 20201002 did some changes.
// Homepage: https://custom-build-robots.com
// Many thanks go to: https://iotdesignpro.com/projects/esp32-cam-surveillance-robot-car
// for doing that greate job to enhance the code to control a robot car.

#include "esp_http_server.h"
#include "esp_timer.h"
#include "esp_camera.h"
#include "img_converters.h"
#include "camera_index.h"
#include "Arduino.h"

// Set the default speed for the left and right motor
// the value is between 0 and 99 like percent.
int speed_left = 0;
int speed_right = 0;

// Those variables are used to control the servo controller
// and to update the OLED display with some status information.
// I knew it is not a really nice way how I programmed that... 
extern String Camerafeed;
extern String MyIP;
extern String message_1;
extern String message_2;
extern int servo_l;
extern int servo_r;
extern int pwmValue1;
extern int pwmValue2;
extern String motor_l;
extern String motor_l_direction;
extern String motor_r;
extern String motor_r_direction;

typedef struct {
        size_t size; //number of values used for filtering
        size_t index; //current value index
        size_t count; //value count
        int sum;
        int * values; //array to be filled with values
} ra_filter_t;

typedef struct {
        httpd_req_t *req;
        size_t len;
} jpg_chunking_t;

#define PART_BOUNDARY "123456789000000000000987654321"
static const char* _STREAM_CONTENT_TYPE = "multipart/x-mixed-replace;boundary=" PART_BOUNDARY;
static const char* _STREAM_BOUNDARY = "\r\n--" PART_BOUNDARY "\r\n";
static const char* _STREAM_PART = "Content-Type: image/jpeg\r\nContent-Length: %u\r\n\r\n";

static ra_filter_t ra_filter;
httpd_handle_t stream_httpd = NULL;
httpd_handle_t camera_httpd = NULL;

static ra_filter_t * ra_filter_init(ra_filter_t * filter, size_t sample_size){
    memset(filter, 0, sizeof(ra_filter_t));

    filter->values = (int *)malloc(sample_size * sizeof(int));
    if(!filter->values){
        return NULL;
    }
    memset(filter->values, 0, sample_size * sizeof(int));

    filter->size = sample_size;
    return filter;
}

static int ra_filter_run(ra_filter_t * filter, int value){
    if(!filter->values){
        return value;
    }
    filter->sum -= filter->values[filter->index];
    filter->values[filter->index] = value;
    filter->sum += filter->values[filter->index];
    filter->index++;
    filter->index = filter->index % filter->size;
    if (filter->count < filter->size) {
        filter->count++;
    }
    return filter->sum / filter->count;
}

static size_t jpg_encode_stream(void * arg, size_t index, const void* data, size_t len){
    jpg_chunking_t *j = (jpg_chunking_t *)arg;
    if(!index){
        j->len = 0;
    }
    if(httpd_resp_send_chunk(j->req, (const char *)data, len) != ESP_OK){
        return 0;
    }
    j->len += len;
    return len;
}

static esp_err_t stream_handler(httpd_req_t *req){
    camera_fb_t * fb = NULL;
    esp_err_t res = ESP_OK;
    size_t _jpg_buf_len = 0;
    uint8_t * _jpg_buf = NULL;
    char * part_buf[64];

    static int64_t last_frame = 0;
    if(!last_frame) {
        last_frame = esp_timer_get_time();
    }

    res = httpd_resp_set_type(req, _STREAM_CONTENT_TYPE);
    if(res != ESP_OK){
        return res;
    }

    while(true){
        fb = esp_camera_fb_get();
        if (!fb) {
            Serial.printf("Camera capture failed");
            res = ESP_FAIL;
        } else {
            if(fb->format != PIXFORMAT_JPEG){
                bool jpeg_converted = frame2jpg(fb, 80, &_jpg_buf, &_jpg_buf_len);
                esp_camera_fb_return(fb);
                fb = NULL;
                if(!jpeg_converted){
                    Serial.printf("JPEG compression failed");
                    res = ESP_FAIL;
                }
            } else {
                _jpg_buf_len = fb->len;
                _jpg_buf = fb->buf;
            }
        }
        if(res == ESP_OK){
            size_t hlen = snprintf((char *)part_buf, 64, _STREAM_PART, _jpg_buf_len);
            res = httpd_resp_send_chunk(req, (const char *)part_buf, hlen);
        }
        if(res == ESP_OK){
            res = httpd_resp_send_chunk(req, (const char *)_jpg_buf, _jpg_buf_len);
        }
        if(res == ESP_OK){
            res = httpd_resp_send_chunk(req, _STREAM_BOUNDARY, strlen(_STREAM_BOUNDARY));
        }
        if(fb){
            esp_camera_fb_return(fb);
            fb = NULL;
            _jpg_buf = NULL;
        } else if(_jpg_buf){
            free(_jpg_buf);
            _jpg_buf = NULL;
        }
        if(res != ESP_OK){
            break;
        }
        int64_t fr_end = esp_timer_get_time();

        int64_t frame_time = fr_end - last_frame;
        last_frame = fr_end;
        frame_time /= 1000;
        uint32_t avg_frame_time = ra_filter_run(&ra_filter, frame_time);
        Serial.printf("MJPG: %uB %ums (%.1ffps), AVG: %ums (%.1ffps)"
            ,(uint32_t)(_jpg_buf_len),
            (uint32_t)frame_time, 1000.0 / (uint32_t)frame_time,
            avg_frame_time, 1000.0 / avg_frame_time
        );
    }

    last_frame = 0;
    return res;
}

static esp_err_t index_handler(httpd_req_t *req){
     httpd_resp_set_type(req, "text/html");
     String page = "";
     page += "<TITLE>Robot Car control center</TITLE>";
     page += "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=0\">\n";
     page += "<script>var xhttp = new XMLHttpRequest();</script>";
     page += "<script>function getsend(arg) { xhttp.open('GET', arg +'?' + new Date().getTime(), true); xhttp.send() } </script>";
     
     page += "<p align=center><IMG SRC='http://" + Camerafeed + ":81/stream' style='width:350px; transform:rotate(0deg);'></p><br/>";
     
     page += "<p align=center> <button style=width:90px;height:30px; onmousedown=getsend('forward')  ontouchstart=getsend('forward')  >^</button> </p>";
     
     page += "<p align=center>";
     page += "<button style=width:90px;height:30px onmousedown=getsend('leftturn')  ontouchstart=getsend('leftturn') ><</button>&nbsp;";
     page += "<button style=width:90px;height:30px onmousedown=getsend('stop') onmouseup=getsend('stop')>STOP</button>&nbsp;";
     page += "<button style=width:90px;height:30px onmousedown=getsend('rightturn')  ontouchstart=getsend('rightturn') >></button>";
     page += "</p>";
    
     page += "<p align=center><button style=width:90px;height:30px onmousedown=getsend('reverse')  ontouchstart=getsend('reverse')  >v</button></p>";  
 
    return httpd_resp_send(req, &page[0], strlen(&page[0]));
}

static esp_err_t forward_handler(httpd_req_t *req){
      
      speed_left += 20;
      speed_right += 20;

      // check that speed_left is not higher then 100%
      if (speed_left >= 100)
      {
        speed_left = 99;
      }
      // check that speed_right is not higher then 100%
      if (speed_right >= 100)
      {
        speed_right = 99;
      }

      // set the turning direction and speed of the left motor
      if (speed_left >=0)
      {

        servo_l = 0;
        pwmValue1 = 4096 + (40.96 * speed_left);
        pwmValue2 = 0;
        motor_l = "left";
        motor_l_direction = "forward";

      }
      else if (speed_left < 0)
      {
        
        servo_l = 0;
        pwmValue1 = 4096 + (40.96 * speed_left);
        pwmValue2 = 0;
        motor_l = "left";
        motor_l_direction = "backward";
                
      }
      
      // set the turning direction and speed of the right motor
      if (speed_right >=0)
      {

        servo_r = 5;
        pwmValue1 = 4096 - (40.96 * speed_right);
        pwmValue2 = 0;
        motor_r = "right";
        motor_r_direction = "forward";

      }
      else if (speed_right < 0)
      {

        servo_r = 5;
        pwmValue1 = 4096 + (40.96 * speed_right);
        pwmValue2 = 0;
        motor_r = "right";
        motor_r_direction = "backward";
                
      } 
        
    message_1 = ("Left speed: "+String(speed_left));
    message_2 = ("Right speed: "+String(speed_right));
    
    httpd_resp_set_type(req, "text/html");
    return httpd_resp_send(req, "OK", 2);
}
static esp_err_t reverse_handler(httpd_req_t *req){

      speed_left -= 20;
      speed_right -= 20;

      // check that speed_left is not higher then 100%
      if (speed_left <= -100)
      {
        speed_left = -99;
      }
      // check that speed_right is not higher then 100%
      if (speed_right <= -100)
      {
        speed_right = -99;
      }

      // set the turning direction and speed of the left motor
      if (speed_left >=0)
      {

        servo_l = 0;
        pwmValue1 = 4096 - (40.96 * speed_left);
        pwmValue2 = 0;
        motor_l = "left";
        motor_l_direction = "forward";
                
      }
      else if (speed_left < 0)
      {

        servo_l = 0;
        pwmValue1 = 4096 + (40.96 * speed_left);
        pwmValue2 = 0;
        motor_l = "left";
        motor_l_direction = "backward";
        
      }
      
      // set the turning direction and speed of the right motor
      if (speed_right >=0)
      {
       
        servo_r = 5;
        pwmValue1 = 4096 - (40.96 * speed_right);
        pwmValue2 = 0;
        motor_r = "right";
        motor_r_direction = "forward";
                
      }
      else if (speed_right < 0)
      {

        servo_r = 5;
        pwmValue1 = 4096 + (40.96 * speed_right);
        pwmValue2 = 0;
        motor_r = "right";
        motor_r_direction = "backward";
        
      }      
          
    message_1 = ("Left speed: "+String(speed_left));
    message_2 = ("Right speed: "+String(speed_right));    

    httpd_resp_set_type(req, "text/html");
    return httpd_resp_send(req, "OK", 2);
}

static esp_err_t leftturn_handler(httpd_req_t *req){

      speed_left += 20;
      speed_right -= 20;

      // check that speed_left is not higher then 100%
      if (speed_left >= 100)
      {
        speed_left = 99;
      }
      // check that speed_right is not higher then 100%
      if (speed_right <= -100)
      {
        speed_right = -99;
      }

      // set the turning direction and speed of the left motor
      if (speed_left >=0)
      {

        servo_l = 0;
        pwmValue1 = 4096 - (40.96 * speed_left);
        pwmValue2 = 0;
        motor_l = "left";
        motor_l_direction = "forward";

      }
      else if (speed_left < 0)
      {

        servo_l = 0;
        pwmValue1 = 4096 + (40.96 * speed_left);
        pwmValue2 = 0;
        motor_l = "left";
        motor_l_direction = "backward";

      }

      // set the turning direction and speed of the right motor
      if (speed_right >=0)
      {

        servo_r = 5;
        pwmValue1 = 4096 - (40.96 * speed_right);
        pwmValue2 = 0;
        motor_r = "right";
        motor_r_direction = "forward";

      }
      else if (speed_right < 0)
      {

        servo_r = 5;
        pwmValue1 = 4096 + (40.96 * speed_right);
        pwmValue2 = 0;
        motor_r = "right";
        motor_r_direction = "backward";

      }                
    
    message_1 = ("Left speed: "+String(speed_left));
    message_2 = ("Right speed: "+String(speed_right));   
    
    httpd_resp_set_type(req, "text/html");
    return httpd_resp_send(req, "OK", 2);
}
static esp_err_t rightturn_handler(httpd_req_t *req){

      speed_left -= 20;
      speed_right += 20;

      // check that speed_left is not higher then 100%
      if (speed_left <= -100)
      {
        speed_left = -99;
      }
      // check that speed_right is not higher then 100%
      if (speed_right >= 100)
      {
        speed_right = 99;
      }

      // set the turning direction and speed of the left motor
      if (speed_left >=0)
      {

        servo_l = 0;
        pwmValue1 = 4096 - (40.96 * speed_left);
        pwmValue2 = 0;
        motor_l = "left";
        motor_l_direction = "forward";

      }
      else if (speed_left < 0)
      {

        servo_l = 0;
        pwmValue1 = 4096 + (40.96 * speed_left);
        pwmValue2 = 0;
        motor_l = "left";
        motor_l_direction = "backward";
        
      }
      
      // set the turning direction and speed of the right motor
      if (speed_right >=0)
      {

        servo_r = 5;
        pwmValue1 = 4096 - (40.96 * speed_right);
        pwmValue2 = 0;
        motor_r = "right";
        motor_r_direction = "forward";

      }
      else if (speed_right < 0)
      {

        servo_r = 5;
        pwmValue1 = 4096 + (40.96 * speed_right);
        pwmValue2 = 0;
        motor_r = "right";
        motor_r_direction = "backward";
   
      }      
          
    message_1 = ("Left speed: "+String(speed_left));
    message_2 = ("Right speed: "+String(speed_right));   
    
    httpd_resp_set_type(req, "text/html");
    return httpd_resp_send(req, "OK", 2);
}

static esp_err_t stop_handler(httpd_req_t *req){

   speed_left = 0;
   speed_right = 0;

   servo_l = 0;
   pwmValue1 = 0;
   pwmValue2 = 4096;
        
   servo_r = 5;
   pwmValue1 = 0;
   pwmValue2 = 4096;
     
   message_1 = ("Left speed: "+String(speed_left));
   message_2 = ("Right speed: "+String(speed_right));   
    
    httpd_resp_set_type(req, "text/html");
    return httpd_resp_send(req, "OK", 2);
}


void startCameraServer(){
    httpd_config_t config = HTTPD_DEFAULT_CONFIG();

    httpd_uri_t forward_uri = {
        .uri       = "/forward",
        .method    = HTTP_GET,
        .handler   = forward_handler,
        .user_ctx  = NULL
    };

    httpd_uri_t reverse_uri = {
        .uri       = "/reverse",
        .method    = HTTP_GET,
        .handler   = reverse_handler,
        .user_ctx  = NULL
    };

    httpd_uri_t stop_uri = {
        .uri       = "/stop",
        .method    = HTTP_GET,
        .handler   = stop_handler,
        .user_ctx  = NULL
    };

    httpd_uri_t leftturn_uri = {
        .uri       = "/leftturn",
        .method    = HTTP_GET,
        .handler   = leftturn_handler,
        .user_ctx  = NULL
    };
    
    httpd_uri_t rightturn_uri = {
        .uri       = "/rightturn",
        .method    = HTTP_GET,
        .handler   = rightturn_handler,
        .user_ctx  = NULL
    };
    
   
    httpd_uri_t index_uri = {
        .uri       = "/",
        .method    = HTTP_GET,
        .handler   = index_handler,
        .user_ctx  = NULL
    };

   httpd_uri_t stream_uri = {
        .uri       = "/stream",
        .method    = HTTP_GET,
        .handler   = stream_handler,
        .user_ctx  = NULL
    };


    ra_filter_init(&ra_filter, 20);
    Serial.printf("Starting web server on port: '%d'", config.server_port);
    if (httpd_start(&camera_httpd, &config) == ESP_OK) {
        httpd_register_uri_handler(camera_httpd, &index_uri);
        httpd_register_uri_handler(camera_httpd, &forward_uri); 
        httpd_register_uri_handler(camera_httpd, &reverse_uri); 
        httpd_register_uri_handler(camera_httpd, &stop_uri); 
        httpd_register_uri_handler(camera_httpd, &leftturn_uri);
        httpd_register_uri_handler(camera_httpd, &rightturn_uri);
       
    }

    config.server_port += 1;
    config.ctrl_port += 1;
    Serial.printf("Starting stream server on port: '%d'", config.server_port);
    if (httpd_start(&stream_httpd, &config) == ESP_OK) {
        httpd_register_uri_handler(stream_httpd, &stream_uri);
    }
}
